#!/usr/bin/ruby
require 'colorize'
require 'yaml'
require 'transacted'
require 'metaverse/repo'
require 'metaverse/iterator'
require 'metaverse/errors'

module Metaverse
  class Base
    def initialize path
      @logger = Logger.new STDOUT
      @base_path = path
      read_config "#{path}/.meta.yml"
      @repos_paths = Metaverse::Iterator.new path, @ignored_repos, @repos_paths
      if @repos_paths.empty?
        @repos_paths.build 
        save_config "#{path}/.meta.yml"
        puts "Repos cache built.".green
      end      
      @repos = @repos_paths.map {|repo| Metaverse::Repo.new repo}
    end

    def status
      check_dirtiness
      check_consistency
    end


    def checkout name
      action_options = -> (repo) {
        {
          up: -> {
            puts "\n # #{repo.name}".blue
            repo.checkout name
          },
          down: -> {
            puts "\n # Rolling back in #{repo.name}".blue
            repo.checkout repo.pop_previous_branch
          }
        }
      }

      if check_dirtiness and check_consistency
        actions = @repos.map {|repo| 
          puts "\n # #{repo.name}".blue
          Transacted::Action.new action_options.call(repo)
        }

        checkout_transaction = Transacted::Transaction.new actions
        case checkout_transaction.execute
        when :execution_success then puts "Checkout successful".green
        when :rollback_success then raise "An error prevented checking out the system. Rolled back successfully".yellow
        when :rollback_failure then raise "An error was encountered during the checkout. In addition, an error happened while trying to rollback the system. Please fix the state of your system manually.".red
        end
      end
        
    end

    def create_state prefix, state
      @repos.each { |repo|
        puts "\n # #{repo.name}".blue
        repo.create_state prefix, state
        repo.checkout "#{prefix}/#{state}" if not prefix == "snapshot"
      } if check_dirtiness and check_consistency
    end


    def load_state prefix, state, remote = nil, should_create_branch = false
      action_options = -> (repo) {
        {
          up: -> { 
            puts "\n # #{repo.name}".blue
            repo.load_state prefix, state, remote, should_create_branch
          },
          down: -> { 
            puts "\n # Rolling back in #{repo.name}".blue
            repo.checkout repo.peek_previous_branch
            repo.delete_branch repo.pop_previous_branch if should_create_branch
          }
        }
      }

      if check_dirtiness and check_consistency
        
        actions = @repos.map { |repo|
          Transacted::Action.new action_options.call repo
        }

        load_state_transaction = Transacted::Transaction.new actions
        case load_state_transaction.execute
        when :execution_success then puts "Loading state successful".green
        when :rollback_success then raise "An error prevented loading the state of the system. Rolled back successfully".yellow
        when :rollback_failure then raise "An error was encountered during the loading of the state. In addition, an error happened while trying to rollback the system. Please fix the state of your system manually.".red
        end
      end
    end


    def send_state prefix, state, remote, should_clean = false
      @repos.each { |repo|
        puts "\n # #{repo.name}".blue
        is_branch = repo.current_branch.match /refs\/heads\/(.*)/
        repo.send_state prefix, state, remote, should_clean
      } if check_dirtiness and check_consistency
    end


    def update remote = nil
      @repos.each { |repo|
        puts "\n # #{repo.name}".blue
        repo.update remote
      } if check_dirtiness and check_consistency
    end


    def branches
      @repos.each { |repo| puts "#{repo.name}@#{repo.current_branch}" }
    end

    def system_state
      branches = Hash.new 0
      @repos.each {|repo| branches[repo.current_branch] += 1}

      state = branches.max{|a,b| a[1] <=> b[1]}
      state[0] = state.first.strip
      state
    end

    def check_dirtiness
      dirty_repos = @repos.reject { |repo|
        !repo.dirty?
      }.map &:name

      if dirty_repos.length == 0
        puts 'The system is clean'.green
      else
        puts 'The system is dirty. Please check the following repos :'.red , dirty_repos
      end
      dirty_repos.length == 0 
    end

    def check_consistency
      branch_name = system_state.first
      inconsistent_repos = @repos.reject { |repo|
        repo.current_branch == branch_name
      }.map &:name

      if inconsistent_repos.length == 0
        puts 'The system is consistent'.green
      else
        puts 'The system is inconsistent. Please check the following repos :'.red , inconsistent_repos
      end
      inconsistent_repos.length == 0 
    end

    def read_config path
      Errors::config_not_found! if not File.exist? path

      config = YAML.load File.open(path)
      @repos_paths = config['repos'] || []
      @origin_remote = config['remotes']['main']
      @own_remote = config['remotes']['own']
      @ignored_repos = config['ignore'] || []
    end

    def save_config path
      config = {
        'repos' => @repos_paths.map{ |repo| repo },
        'remotes'=> {
          'main' => @origin_remote,
          'own' => @own_remote
        },
        'ignore' => @ignored_repos
      }
      File.open(path, 'w') {|f| f.write config.to_yaml }
    end

    def clear_repos
      @repos_paths = []
      save_config "#{@base_path}/.meta.yml"
      puts "Repos cache cleared.".green
    end

    def pull
      @repos.each { |repo|
        puts "\n # #{repo.name}".blue
        repo.pull @origin_remote
      } if check_dirtiness and check_consistency
    end

    def add_remote name, base_url
      @repos.each { |repo|
        puts "\n # #{repo.name}".blue
        repo.add_remote name, "#{base_url}/#{repo.name}"
      }
    end

    def exec env, command
      @repos.each { |repo|
        puts "\n # #{repo.name}".blue
        repo.exec env, command
      }
    end
  end
end