require 'rugged'
require 'iniparse'

module Metaverse
  class Repo

    # Initialize a git repository 
    # path - An absolute path to the repo root folder

    def initialize path
      load path
      @previous_branches = []
    end


    def load path
      @repo = Rugged::Repository.new(path)
      @repo.remotes.each_name do |remote|
        fetch_refspec = "+refs/meta/local/*:refs/meta/remotes/#{remote}/*"
        push_refspec = "+refs/meta/local/*:refs/meta/local/*"
        add_option_to_remote remote, fetch_refspec, 'fetch'
        add_option_to_remote remote, push_refspec, 'push'
      end
    end


    def name
      File.basename @repo.workdir
    end

    def workdir
      @repo.workdir
    end

    def checkout ref
      if @repo.branches[ref].nil?
        is_reference = false 
        if Rugged::Reference.valid_name?(ref) and @repo.references[ref].nil?
          raise Errors::ref_not_found ref
        end
      end

      @previous_branches.push current_branch
      Dir.chdir(@repo.workdir) do
        `git checkout #{ref}`#TODO: find out what's going on, check out a branch
      end
    end


    def current_branch
      branch_name = ''

      Dir.chdir(@repo.workdir) do
        branch_name = `git branch | sed -n '/\* /s///p'`
        #TODO Look into why the command returns an empty string in a new repo
        if branch_name == ''
          branch_name = 'master'
        end
      end

      if branch_name.include? "detached"
        match = /[A-Za-z0-9._-]+(?:\/[A-Za-z0-9._-]+)+/.match(branch_name) 
        if not match.nil?
          return match[0]
        end
        return 'HEAD'
      end

      branch_name.strip
    end

    def peek_previous_branch
      @previous_branches.last
    end

    def pop_previous_branch
      @previous_branches.pop
    end

    def head
      @repo.head
    end


    def branches
      @repo.branches
    end


    def create_ref ref_name, ref_target
      @repo.references.create ref_name, ref_target.target_id
    end


    def update_ref ref_name, ref_target
      @repo.references.update ref_name, ref_target.target_id
    end


    def create_state prefix, state
      ref_target = @repo.head
      if not prefix == 'snapshot'
        branch_name = "#{prefix}/#{state}"
        ref_target = @repo.create_branch branch_name
      end
      ref_name = "refs/meta/local/#{prefix}/#{state}"
      create_ref ref_name, ref_target
    end


    def load_state prefix, state, remote = nil, should_create_branch = false
      ref_name = "refs/meta"
      ref_name += remote.nil? ? "/local" : "/remotes/#{remote}"
      ref_name += "/#{prefix}/#{state}"
      if should_create_branch
        branch_name = "#{prefix}/#{state}"
        # TODO: Decide what to do when a branch exists previous. Should we :
        # - Replace it's ref by the saved state ? ( We lose the previous branch reference)
        # - Load it as it is ( We may get an inconsistent state since the branch could point to a different commit than the state reference )
        if @repo.branches[branch_name].nil?
          puts "Creating branch #{branch_name} based on #{ref_name}"
          @repo.create_branch branch_name, ref_name
        end
        return checkout branch_name
      end
      checkout ref_name
    end


    def send_state prefix, state, remote, should_clean = false
      refs = ["refs/meta/local/#{prefix}/#{state}"]
      branch_name = "#{prefix}/#{state}"
      branch = @repo.branches[branch_name]
      develop = @repo.branches['develop']

      if branch.nil?
        Errors::ref_not_found branch_name
        return false
      end

      if develop.nil?
        Errors::ref_not_found 'develop'
        return false
      end

      if (not prefix == 'snapshot') and ahead_of_develop? branch
          refs << branch_name
          update_ref refs[0], @repo.branches[branch_name]
      end

      clean_branch branch if should_clean

      Dir.chdir(@repo.workdir) do
        `git push #{remote} #{refs.join(' ')}`
      end
    end


    def clean_branch branch

      changed = ahead_of_develop? branch
      if not changed
        checkout 'develop'
        @repo.branches.delete branch
      end
      changed
    end

    def delete_branch branch
      if not @repo.branches[branch].nil?
        @repo.branches.delete branch
      end
    end

    def dirty?
      is_dirty = false
      @repo.status { |file, data|
        next if data == [:ignored]
        is_dirty = true
      }
      is_dirty
    end


    def update remote = nil
      if not remote.nil? and @repo.remotes[remote].nil?
        return Errors::remote_not_found remote
      end
      Dir.chdir(@repo.workdir) do
        `git remote update #{remote}`
      end
    end


    def add_option_to_remote remote, value, key
      config_path = "#{@repo.path}config"
      cfg = IniParse.parse(File.read(config_path))
      if not [*cfg["remote \"#{remote}\""][key]].include? value
        cfg["remote \"#{remote}\""][key] = [*cfg["remote \"#{remote}\""][key], value]
        cfg.save config_path
      end
    end


    def has_changes? ref1, ref2
      not ref1.target_id == ref2.target_id
    end


    def ahead_of_develop? branch
      has_changes? branch, @repo.branches['develop']
    end

    def pull remote
      if @repo.remotes[remote].nil?
        return Errors::remote_not_found remote
      end
      Dir.chdir(@repo.workdir) do
        `git pull #{remote} #{current_branch}`
      end
    end

    def add_remote name, url
      return Errors::remote_exists name if not @repo.remotes[name].nil?
      @repo.remotes.create name, url
    end

    def exec env, command
      Dir.chdir(@repo.workdir) do
        system env, command
      end
    end
  end
end